@@ -2,7 +2,7 @@ package Net::DNS::RR::SVCB;
 
 use strict;
 use warnings;
-our $VERSION = (qw$Id: SVCB.pm 1832 2021-03-22 08:33:36Z willem $)[2];
+our $VERSION = (qw$Id: SVCB.pm 1843 2021-07-09 09:10:08Z willem $)[2];
 
 use base qw(Net::DNS::RR);
 
@@ -23,6 +23,17 @@ use Net::DNS::RR::AAAA;
 use Net::DNS::Text;
 
 
+my %keybyname = (
+	mandatory	  => 'key0',
+	alpn		  => 'key1',
+	'no-default-alpn' => 'key2',
+	port		  => 'key3',
+	ipv4hint	  => 'key4',
+	ech		  => 'key5',
+	ipv6hint	  => 'key6',
+	);
+
+
 sub _decode_rdata {			## decode rdata from wire-format octet string
 	my $self = shift;
 	my ( $data, $offset ) = @_;
@@ -33,15 +44,14 @@ sub _decode_rdata {			## decode rdata from wire-format octet string
 	my $index;
 	( $self->{TargetName}, $index ) = Net::DNS::DomainName->decode( \$rdata, 2 );
 
-	my $params = $self->{SvcParams} = {};
-	my $limit  = length($rdata) - 4;
+	my $params = $self->{SvcParams} = [];
+	my $limit  = length($rdata) - 3;
 	while ( $index < $limit ) {
 		my ( $key, $size ) = unpack( "\@$index n2", $rdata );
-		$params->{$key} = substr $rdata, $index + 4, $size;
+		push @$params, ( $key, substr $rdata, $index + 4, $size );
 		$index += ( $size + 4 );
 	}
-	die 'corrupt RDATA in ' . $self->type unless $index == $self->{rdlength};
-	$self->_post_parse;
+	die $self->type . ': corrupt RDATA' unless $index == length($rdata);
 	return;
 }
 
@@ -49,23 +59,24 @@ sub _decode_rdata {			## decode rdata from wire-format octet string
 sub _encode_rdata {			## encode rdata as wire-format octet string
 	my $self = shift;
 
-	my ($params) = grep {defined} $self->{SvcParams}, {};
-	my @params;
-	foreach ( sort { $a <=> $b } keys %$params ) {
-		my $value = $params->{$_};
-		next unless defined $value;
-		push @params, pack( 'n2a*', $_, length($value), $value );
+	my @packed = pack 'n a*', $self->{SvcPriority}, $self->{TargetName}->encode;
+	my $params = $self->{SvcParams} || [];
+	my @params = @$params;
+	while (@params) {
+		my $key = shift @params;
+		my $val = shift @params;
+		push @packed, pack( 'n2a*', $key, length($val), $val );
 	}
-	return pack 'n a* a*', $self->{SvcPriority}, $self->{TargetName}->encode, join '', @params;
+	return join '', @packed;
 }
 
 
 sub _format_rdata {			## format rdata portion of RR string.
 	my $self = shift;
 
-	my ($params) = grep {defined} $self->{SvcParams}, {};
 	my $priority = $self->{SvcPriority};
-	return ( $priority, $self->{TargetName}->string ) unless scalar keys %$params;
+	my $params   = $self->{SvcParams} || [];
+	return ( $priority, $self->{TargetName}->string ) unless scalar @$params;
 
 	my @rdata  = unpack 'H4', pack 'n', $priority;
 	my $target = $self->{TargetName}->encode();
@@ -74,14 +85,15 @@ sub _format_rdata {			## format rdata portion of RR string.
 	push @rdata, $length > 18 ? "\t; $priority\n" : (), @target;
 	push @rdata, join '', "\t; ", $self->{TargetName}->string, "\n" if $length > 3;
 
-	foreach ( sort { $a <=> $b } keys %$params ) {
-		my $value = $params->{$_};
-		next unless defined $value;
+	my @params = @$params;
+	while (@params) {
+		my $key = shift @params;
+		my $val = shift @params;
 		push @rdata, "\n";
-		push @rdata, "; key$_=...\n" if $_ > 15;
-		push @rdata, unpack 'H4H4', pack( 'n2', $_, length $value );
-		push @rdata, split /(\S{32})/, unpack 'H*', $value;
-		$length += 4 + length $value;
+		push @rdata, "; key$key=...\n" if $key > 15;
+		push @rdata, unpack 'H4H4', pack( 'n2', $key, length $val );
+		push @rdata, split /(\S{32})/, unpack 'H*', $val;
+		$length += 4 + length $val;
 	}
 	return ( "\\# $length", @rdata );
 }
@@ -93,20 +105,22 @@ sub _parse_rdata {			## populate RR from rdata in argument list
 	$self->svcpriority(shift);
 	$self->targetname(shift);
 
-	while ( my $attribute = shift ) {
-		for ($attribute) {
-			my @argument = '';
+	local $SIG{__WARN__} = sub { die @_ };
+	while ( my $svcparam = shift ) {
+		for ($svcparam) {
+			my @value;
 			if (/=(.*)$/) {
-				for ( my $rhs = length($1) ? $1 : shift ) {
-					s/^(["'])(.*)\1$/$2/;	# strip paired quotes
-					s/\\,/\\044/g;		# disguise escaped comma
-					@argument = split /,/;	# potentially multi-valued
-				}
+				local $_ = length($1) ? $1 : shift;
+				s/^(["'])(.*)\1$/$2/;		# strip paired quotes
+				s/\\,/\\044/g;			# disguise escaped comma
+				push @value, split /,/;
+			} else {
+				push @value, '' unless $keybyname{lc $_};    # empty keyNNN
 			}
 
-			s/[-]/_/g;				# extract attribute identifier
+			s/[-]/_/g;				# extract identifier
 			m/^([^=]+)/;
-			$self->$1(@argument);
+			$self->$1(@value);
 		}
 	}
 	return;
@@ -117,13 +131,18 @@ sub _post_parse {			## parser post processing
 	my $self = shift;
 
 	my $params = $self->{SvcParams} || return;
-	my %unique;
-	my @unique = grep { !$unique{$_}++ } unpack 'n*', ( $params->{0} || return );
-	croak( $self->type . qq(: mandatory "key0" not permitted) ) if $unique{0};
-	foreach (@unique) {
-		croak( $self->type . qq(: duplicate "key$_" in mandatory list) ) if --$unique{$_};
-		croak( $self->type . qq(: mandatory "key$_" not defined) )	 if !defined( $params->{$_} );
+	my %params = @$params;
+	$self->key0(undef);					# ruse to force sorting of SvcParams
+	if ( defined $params{0} ) {
+		my %unique;
+		foreach ( grep { !$unique{$_}++ } unpack 'n*', $params{0} ) {
+			croak( $self->type . qq[: unexpected "key0" in mandatory list] ) if $unique{0};
+			croak( $self->type . qq[: duplicate "key$_" in mandatory list] ) if --$unique{$_};
+			croak( $self->type . qq[: mandatory "key$_" not present] ) unless defined $params{$_};
+		}
+		$self->mandatory( keys %unique );		# restore mandatory key list
 	}
+	croak( $self->type . qq[: expected alpn="..." not present] ) if defined( $params{2} ) and !$params{1};
 	return;
 }
 
@@ -157,80 +176,53 @@ sub targetname {
 
 ########################################
 
-use constant ASCII => chr(65) eq 'A';
 
-my %escape = do {			## precalculated ASCII escape table
-	my @escape = ( 0 .. 32, 34, 92, 127 .. 255 );		# numerical escape
-
-	my %table = map { ( chr($_) => chr($_) ) } ( 0 .. 255 );
-
-	foreach my $codepoint (@escape) {
-		my $ddd = sprintf( '%03u', $codepoint );
-		$ddd =~ tr [0-9] [\060-\071];			# transliterate non-ASCII
-		$table{pack( 'C', $codepoint )} = pack 'C a3', 92, $ddd;
-	}
-
-	%table;
-};
-
-sub _raw {				## concatenate @_ and escape non-printable
+sub _presentation {			## render octet string(s) in presentation format
 	return () unless scalar @_;
-	local $_ = join '', map { $escape{$_} } map { split( //, $_ ) } @_;
-
-	# partial transliteration for non-ASCII character encodings
-	tr
-	[\040-\176\000-\377]
-	[ !"#$%&'()*+,\-./0-9:;<=>?@A-Z\[\\\]^_`a-z{|}~?] unless ASCII;
-
-	return $_;
+	my $raw = join '', @_;
+	my $txt = Net::DNS::Text->decode( \$raw, 0, length($raw) );
+	return map { s/ /\\032/g; $_ } $txt->string;
 }
 
 sub _base64 {
-	return _raw( map { MIME::Base64::decode($_) } @_ );
+	return _presentation( map { MIME::Base64::decode($_) } @_ );
 }
 
 sub _integer16 {
-	return _raw( map { pack( 'n', $_ ) } @_ );
+	return _presentation( map { pack( 'n', $_ ) } @_ );
 }
 
 sub _ipv4 {
-	return _raw( map { bless( {}, 'Net::DNS::RR::A' )->address($_) } @_ );
+	return _presentation( map { Net::DNS::RR::A::address( {}, $_ ) } @_ );
 }
 
 sub _ipv6 {
-	return _raw( map { bless( {}, 'Net::DNS::RR::AAAA' )->address($_) } @_ );
+	return _presentation( map { Net::DNS::RR::AAAA::address( {}, $_ ) } @_ );
 }
 
 sub _string {
-	return _raw( map { Net::DNS::Text->new($_)->encode() } @_ );
+	return _presentation( map { Net::DNS::Text->new($_)->encode() } @_ );
 }
 
 
-my %keybyname = (
-	mandatory	  => 0,
-	alpn		  => 1,
-	'no-default-alpn' => 2,
-	port		  => 3,
-	ipv4hint	  => 4,
-	ech		  => 5,
-	echconfig	  => 5,					# draft-ietf-dnsop-svcb-https compatible
-	ipv6hint	  => 6,
-	);
-
-
 sub mandatory {				## mandatory=key1,port,...
-	my ( $self, @arg ) = grep {defined} @_;
-	my @keys = map { $keybyname{lc $_} || ( /^key(\d+)$/i ? $1 : croak qq["$_" unknown] ) } @arg;
+	my $self = shift;
+	my @list = map { $keybyname{lc $_} || $_ } map { split /,/ } @_;
+	my @keys = map { /(\d+)$/ ? $1 : croak( $self->type . qq[: unexpected "$_"] ) } @list;
 	return $self->key0( _integer16( sort { $a <=> $b } @keys ) );
 }
 
 sub alpn {				## alpn=h3,h2,...
 	my $self = shift;
-	return $self->key1( _string(@_) );
+
+	###	tolerate unnecessary double-escape nonsense in draft-ietf-dnsop-svcb-https	###
+	my @sanitized = map { s/\\092,/\\044/g; s/\\092\\092/\\092/g; split /,/ } join ',', @_;
+	return $self->key1( scalar(@_) ? _string(@sanitized) : () );
 }
 
 sub no_default_alpn {			## no-default-alpn
-	return shift->key2( map {''} @_ );			# uncoverable pod
+	my $self = shift;					# uncoverable pod
+	return $self->key2( ( defined(wantarray) ? @_ : '' ), @_ );
 }
 
 sub port {				## port=1234
@@ -247,7 +239,6 @@ sub ech {				## ech=base64string
 	my $self = shift;
 	return $self->key5( map { _base64($_) } @_ );
 }
-sub echconfig { return &ech; }					# uncoverable pod
 
 sub ipv6hint {				## ipv6hint=2001:DB8::1,...
 	my $self = shift;
@@ -263,21 +254,25 @@ sub AUTOLOAD {				## Dynamic constructor/accessor methods
 	my ($method) = reverse split /::/, $AUTOLOAD;
 
 	my $default = join '::', 'SUPER', $method;
-	return $self->$default(@_) unless $method =~ /^key(\d+)$/i;
+	return $self->$default(@_) unless $method =~ /^key[0]*(\d+)$/i;
 	my $key = $1;
 
-	my ($params) = grep {defined} $self->{SvcParams}, {};
+	my $params = $self->{SvcParams} || [];
+	my %params = @$params;
 
 	if ( scalar @_ ) {
 		my $arg = shift;				# keyNN($value);
-		croak 'unexpected number of arguments' if scalar @_;
-
-		$params->{$key} = defined($arg) ? Net::DNS::Text->new($arg)->raw : undef;
-		$self->{SvcParams} = $params;
+		delete $params{$key} unless defined $arg;
+		croak( $self->type . qq[: duplicate SvcParam "key$key"] ) if defined $params{$key};
+		$params{$key} = Net::DNS::Text->new("$arg")->raw if defined $arg;
+		$self->{SvcParams} = [map { ( $_, $params{$_} ) } sort { $a <=> $b } keys %params];
+		croak( $self->type . qq[: unexpected number of values for "key$key"] ) if scalar @_;
+	} else {
+		croak( $self->type . qq[: no value specified for "key$key"] ) unless defined wantarray;
 	}
 
-	my $value = $params->{$key};
-	return defined($value) ? _raw($value) : $value;
+	my $value = $params{$key};
+	return defined($value) ? _presentation($value) : $value;
 }
 
 
@@ -338,7 +333,7 @@ owner name of this record must be used as the effective TargetName.
     $string = $rr->port();	# \004\210
     $rr->key3($string);
 
-Constructor methods for mnemonic SvcParams defined in draft-ietf-dnsop-svcb-https-02.
+Constructor methods for mnemonic SvcParams defined in draft-ietf-dnsop-svcb-https.
 When invoked without arguments, the methods return the presentation format
 value for the underlying key.
 The behaviour with undefined arguments is not specified.
@@ -347,17 +342,18 @@ The behaviour with undefined arguments is not specified.
 
     $keynn = $rr->keyNN;
     $rr->keyNN( $keynn );
+    $rr->keyNN( undef );
 
 Generic constructor and accessor methods for SvcParams.
 The key index NN is a decimal integer in the range 0 .. 65534.
 The method argument and returned value are both presentation format strings.
 The method returns the undefined value if the key is not present.
-A (key,value) pair will be ignored if the value is undefined.
+The specified key will be deleted if the value is undefined.
 
 
 =head1 COPYRIGHT
 
-Copyright (c)2020 Dick Franks. 
+Copyright (c)2020-2021 Dick Franks. 
 
 All rights reserved.
 
@@ -385,6 +381,6 @@ DEALINGS IN THE SOFTWARE.
 
 =head1 SEE ALSO
 
-L<perl>, L<Net::DNS>, L<Net::DNS::RR>, draft-ietf-dnsop-svcb-https-02
+L<perl>, L<Net::DNS>, L<Net::DNS::RR>, draft-ietf-dnsop-svcb-https
 
 =cut
